import {
  DependencyList,
  useCallback,
  useContext,
  useEffect,
  useMemo,
  useRef,
  useState
} from 'react';
import {
  CheckRequestResult, checkRequestState,
  CheckRequestStateFn,
  IRequestState,
  IRequestStoreOptions,
  RequestStore,
} from '@proscom/prostore';
import { useConnectStore } from './useConnectStore';
import { useContextStore } from './ProstoreContext';
import { ProstoreSsrContext } from './ProstoreSsrContext';

export interface UseRequestStoreResult<Vars, Data, Options> {
  check: CheckRequestResult;
  state: IRequestState<Vars, Data>;
  load: (options?: Options) => Promise<IRequestState<Vars, Data>>;
  store: RequestStore<Vars, Data, Options>;
}

export interface UseRequestOptions<Vars, Data> {
  checkRequest?: CheckRequestStateFn<Vars, Data>;
}

/**
 * Attached component to the provided store and requests data based on the variables and options
 *
 * @param storeOrName - RequestStore to subscribe to, or its name for it to be injected from the context
 * @param variables - Request variables. If they change, request will be fired again
 * @param options - Additional request options. They are not observed
 * @param hookOptions - Additional hook options:
 *  - checkRequest - Optional function which may be used to override default behavior of determining
 *                   when to fire the request
 */
export function useRequestStore<
  Vars,
  Data,
  Options extends IRequestStoreOptions<Vars, Data>
>(
  storeOrName: string | RequestStore<Vars, Data, Options>,
  variables: Vars,
  options: Options = undefined,
  hookOptions: UseRequestOptions<Vars, Data> = {}
): UseRequestStoreResult<Vars, Data, Options> {
  const checkRequest = hookOptions.checkRequest || checkRequestState;

  // Inject store from context
  const store = useContextStore<RequestStore<Vars, Data, Options>>(storeOrName);

  // Local copy of store state
  const [state, setState] = useState(store.state);

  // Subscribe to store changes
  useConnectStore(store, setState);

  // Check if we need to request data
  const queryCheck = checkRequest(state, variables);

  // Memoized callback to request data
  const funcsRef = useRef<{
    loadData?: UseRequestStoreResult<Vars, Data, Options>['load'];
  }>({});
  funcsRef.current.loadData = (newOptions = undefined) => {
    return store.loadData(variables, newOptions || options);
  };
  const loadData = useCallback((newOptions = undefined) => {
    return funcsRef.current.loadData(newOptions);
  }, []);

  // Request the data if it is should be requested
  useEffect(() => {
    if (queryCheck.request) {
      // Error should not happen here, because
      // it is handled inside RequestStore,
      // but we attach the catch listener just in case
      loadData().catch((e) => console.error(e));
    }
  });

  // Attach ssrContext
  const ssrContext = useContext(ProstoreSsrContext);

  // Perform request if server rendering
  if (store.ssrId && ssrContext && ssrContext.isServer && queryCheck.request) {
    loadData().catch((e) => console.error(e));
  }

  // Return stuff which may be useful in the component
  return { check: queryCheck, state, load: loadData, store };
}

/**
 * The same as useRequestStore, but it uses variableCreator which is supplied to useMemo and
 * can be used to memoize query variables. This may be used when the default behavior of deeply
 * comparing variables is not preferable (i.e., it is not fast enough, or variables are not simple objects)
 *
 * @returns {{load, check, state}} - Current request state
 */
export function useRequestStoreVars<
  Vars,
  Data,
  Options extends IRequestStoreOptions<Vars, Data>
>(
  storeOrName: string | RequestStore<Vars, Data, Options>,
  variableCreator: () => Vars,
  args: DependencyList
) {
  const variables = useMemo(variableCreator, args);
  return useRequestStore(storeOrName, variables);
}
